/**
 * @file problems.hpp
 * @author Ryan Curtin
 *
 * Include each of the test problems for ensmallen.
 */
#ifndef ENSMALLEN_PROBLEMS_PROBLEMS_HPP
#define ENSMALLEN_PROBLEMS_PROBLEMS_HPP

#include "ackley_function.hpp"
#include "aug_lagrangian_test_functions.hpp"
#include "beale_function.hpp"
#include "booth_function.hpp"
#include "bukin_function.hpp"
#include "colville_function.hpp"
#include "cross_in_tray_function.hpp"
#include "drop_wave_function.hpp"
#include "easom_function.hpp"
#include "eggholder_function.hpp"
#include "fw_test_function.hpp"
#include "generalized_rosenbrock_function.hpp"
#include "goldstein_price_function.hpp"
#include "gradient_descent_test_function.hpp"
#include "himmelblau_function.hpp"
#include "holder_table_function.hpp"
#include "levy_function_n13.hpp"
#include "logistic_regression_function.hpp"
#include "matyas_function.hpp"
#include "mc_cormick_function.hpp"
#include "rastrigin_function.hpp"
#include "rosenbrock_function.hpp"
#include "rosenbrock_wood_function.hpp"
#include "schaffer_function_n2.hpp"
#include "schaffer_function_n4.hpp"
#include "schwefel_function.hpp"
#include "sgd_test_function.hpp"
#include "softmax_regression_function.hpp"
#include "sparse_test_function.hpp"
#include "sphere_function.hpp"
#include "styblinski_tang_function.hpp"
#include "three_hump_camel_function.hpp"
#include "wood_function.hpp"

#endif
